"""
:module: pynetworking.sides.client
:synopsis: import everything and setup everything for server side code
:author: Julian Sobott
"""
from pynetworking.common import *

global_data.IS_CLIENT = True
